###############################################################################
# Caleydo - Visualization for Molecular Biology - http://caleydo.org
# Copyright (c) The Caleydo Team. All rights reserved.
# Licensed under the new BSD license, available at http://caleydo.org/license
###############################################################################


from builtins import range
from builtins import object


class ApplicationProxy(object):
  """
  helper class for different applications defined by a namespace and a loader function
  """
  def __init__(self, namespace, loader):
    self.namespace = namespace
    # number of suburls to pop
    self.peeks = namespace.count('/')
    self._loader = loader
    self.app = None

  def init(self):
    if self.app is None:
      self.app = self._loader()
    return self

  def match(self, path):
    # start of a suburl or the whole one
    return path.startswith(self.namespace + '/') or path == self.namespace


class PathDispatcher(object):
  """
  helper class to select an application by path
  """
  def __init__(self, default_app, applications):
    self.default_app = default_app

    self.applications = [ApplicationProxy(key, value) for key, value in applications.items()]
    # print self.applications
    from threading import Lock
    self.lock = Lock()

  def get_application(self, path):
    with self.lock:
      for app in self.applications:
        if app.match(path):
          return app.init()

  def __call__(self, environ, start_response):
    from werkzeug.wsgi import pop_path_info, get_path_info

    app = self.get_application(get_path_info(environ))
    if app is not None:
      for _ in range(app.peeks):
        pop_path_info(environ)
      app = app.app
      # print get_path_info(environ), app
    else:  # use default app
      app = self.default_app
    return app(environ, start_response)



###############################################################################
# Caleydo - Visualization for Molecular Biology - http://caleydo.org
# Copyright (c) The Caleydo Team. All rights reserved.
# Licensed under the new BSD license, available at http://caleydo.org/license
###############################################################################


from builtins import range
from builtins import object
import numpy as np
import numpy.ma as ma
import datetime as dt
import decimal


class NumpyTablesEncoder(object):
  def __contains__(self, obj):
    if isinstance(obj, np.ndarray):
      return True
    if isinstance(obj, bytes):
      return True
    if isinstance(obj, np.generic):
      return True
    if isinstance(obj, dt.datetime):
      return True
    if isinstance(obj, decimal.Decimal):
      return True
    return False

  def __call__(self, obj, base_encoder):
    if isinstance(obj, np.ndarray):
      if obj.ndim == 1:
        return [base_encoder.default(x) for x in obj]
      else:
        return [base_encoder.default(obj[i]) for i in range(obj.shape[0])]
    if isinstance(obj, np.generic):
      a = np.asscalar(obj)
      if (isinstance(a, float) and np.isnan(a)) or ma.is_masked(a):
        return None
      return a
    if isinstance(obj, dt.datetime):
      return int(obj.timestamp() * 1000)
    if isinstance(obj, decimal.Decimal):
      return float(obj)
    if isinstance(obj, bytes):
      return obj.decode('utf-8')
    return None


n = NumpyTablesEncoder()


def create():
  return n



###############################################################################
# Caleydo - Visualization for Molecular Biology - http://caleydo.org
# Copyright (c) The Caleydo Team. All rights reserved.
# Licensed under the new BSD license, available at http://caleydo.org/license
###############################################################################


from builtins import object, enumerate
import json
import os
from .config import view
from codecs import open
import logging

_log = logging.getLogger(__name__)


def merge(idtype, species):
  if not species:
    return idtype
  return idtype + '_' + species


class FileMapper(object):
  def __init__(self, desc, plugin):
    folder = plugin.folder + '/data/' if not hasattr(plugin, 'inplace') else plugin.folder
    self._path = os.path.join(folder, desc['path'])
    self._map = None
    self._desc = desc
    self.from_idtype = merge(desc['from'], desc.get('fromSpecies'))
    self.to_idtype = merge(desc['to'], desc.get('toSpecies'))
    _log.info('loading csv mapping table from %s to %s', self.from_idtype, self.to_idtype)
    pass

  def _load(self):
    import csv
    import io

    _log.info('loading real mapping file from %s to %s', self.from_idtype, self.to_idtype)
    with io.open(self._path, 'r', newline='', encoding=self._desc.get('encoding', 'utf-8')) as csvfile:
      reader = csv.reader(csvfile, delimiter=self._desc.get('separator', ','),
                          quotechar=str(self._desc.get('quotechar', '"')))
      self._map = dict()
      for i, line in enumerate(reader):
        if i == 0 and self._desc.get('header', False):
          continue
        key = line[0]
        values = line[1].split(self._desc.get('multi', ';'))
        self._map[key] = values

  def __call__(self, ids):
    if not self._map:
      self._load()
    return [self._map.get(id, None) for id in ids]


def to_files(plugins):
  for plugin in plugins:
    index = os.path.join(plugin.folder + '/data/' if not hasattr(plugin, 'inplace') else plugin.folder, 'mapping.json')
    if not os.path.isfile(index):
      continue
    with open(index, 'r') as f:
      desc = json.load(f)
      for di in desc:
        yield FileMapper(di, plugin)


class DataPlugin(object):
  def __init__(self, folder):
    # add a magic plugin for the static data dir
    self.inplace = True  # avoid adding the data suffix
    self.folder = folder
    self.id = os.path.basename(folder)


class StaticFileProvider(object):
  def __init__(self, plugins):
    self.files = list(to_files(plugins))

    cc = view('phovea_server')
    data_plugin = DataPlugin(os.path.join(cc.dataDir, 'data'))
    self.files.extend(to_files([data_plugin]))
    import glob
    extras = [DataPlugin(f) for f in (os.path.dirname(f) for f in glob.glob(cc.dataDir + '/*/mapping.json')) if
              os.path.basename(f) != 'data']
    self.files.extend(to_files(extras))

  def __iter__(self):
    return iter(((f.from_idtype, f.to_idtype, f) for f in self.files))


def create():
  """
  entry point of this plugin
  """
  from .plugin import plugins
  return StaticFileProvider(plugins())



"""
This encoder is required to handle changes of data types in Python 3.7 by adding list() to set().
"""


class SetEncoder(object):
  def __contains__(self, obj):
    return isinstance(obj, set)

  def __call__(self, obj, base_encoder):
    return list(obj)


encoder = SetEncoder()


def create():
  return encoder



###############################################################################
# Caleydo - Visualization for Molecular Biology - http://caleydo.org
# Copyright (c) The Caleydo Team. All rights reserved.
# Licensed under the new BSD license, available at http://caleydo.org/license
###############################################################################


from builtins import object
import os
import jsoncfg
import codecs
from ._utils import replace_nested_variables, extend
import logging

_log = logging.getLogger(__name__)
_c = None
_pre_merge_changes = {}


# method to control call of _init_config()
def _initialize():
  _init_config()


def get(item, section=None, default=None):
  # use global _c
  global _c
  key = item if section is None else section + '.' + item
  keys = key.split('.')
  act = _c
  for p in keys[0:-1]:
    act = act.get(p, {})
  last_key = keys[len(keys) - 1]

  if last_key not in act:
    # prop does not exist
    return default

  v = act[last_key]

  def resolve(x):
    if '.' in x:  # assume absolute
      return get(x)
    else:
      return act.get(x, None)

  v = replace_nested_variables(v, resolve)
  # print key, v

  return v


def set(item, value, section=None):
  global _c, _pre_merge_changes
  key = item if section is None else section + '.' + item
  keys = key.split('.')
  act = _c
  for p in keys[0:-1]:
    if p not in act:
      act[p] = {}
    act = act[p]
  act[keys[len(keys) - 1]] = value

  if _pre_merge_changes is not None:
    # keep track of changes that were done before the recreation of the configuration
    act = _pre_merge_changes
    for p in keys[0:-1]:
      if p not in act:
        act[p] = {}
      act = act[p]
    act[keys[len(keys) - 1]] = value


def getint(item, section=None, default=0):
  return int(get(item, section, default))


def getfloat(item, section=None, default=0.):
  return float(get(item, section, default))


def getboolean(item, section=None, default=False):
  return bool(get(item, section, default))


def getlist(item, section=None, separator='\n', default=[]):
  v = get(item, section, default)
  return v if isinstance(v, list) else v.split(separator)


def view(section):
  return CaleydoConfigSection(section)


class CaleydoConfigSection(object):
  def __init__(self, section):
    self._section = section

  def __getattr__(self, item):
    return self.get(item)

  def __getitem__(self, item):
    return self.get(item)

  def _expand(self, section=None):
    if self._section is None:
      return section
    return self._section + ('.' + section if section is not None else '')

  def get(self, item, section=None, default=None):
    return get(item, self._expand(section), default=default)

  def set(self, item, value, section=None):
    return set(item, value, self._expand(section))

  def getint(self, item, section=None, default=0):
    return getint(item, self._expand(section), default=default)

  def getfloat(self, item, section=None, default=0.):
    return getfloat(item, self._expand(section), default=default)

  def getboolean(self, item, section=None, default=False):
    return getboolean(item, self._expand(section), default=default)

  def getlist(self, item, section=None, separator='\n', default=[]):
    return getlist(item, self._expand(section), separator, default=default)

  def view(self, section):
    return CaleydoConfigSection(self._expand(section))


# add global_config as parameter and call extend() in return statement
def _merge_config(global_config, config_file, plugin_id):
  _log.info(plugin_id)
  # check initialization
  if global_config is None:
    _initialize()
  with codecs.open(config_file, 'r', 'utf-8') as fi:
    c = jsoncfg.loads(fi.read())
  return extend(global_config, {plugin_id: c})


def _init_config():
  from .__init__ import phovea_config
  f = phovea_config()
  global _c
  if _c is not None:
    return
  print('init config')
  _c = {}
  _c = _merge_config(_c, f, 'phovea_server')
  global_ = os.path.abspath(os.environ.get('PHOVEA_CONFIG_PATH', 'config.json'))
  if os.path.exists(global_) and global_ != f:
    print(('configuration file: ' + global_))
    with codecs.open(global_, 'r', 'utf-8') as fi:
      extend(_c, jsoncfg.loads(fi.read()))


def merge_plugin_configs(plugins):
  # merge all the plugins
  global _c, _pre_merge_changes
  # check initialization
  if _c is None:
    _initialize()
  for plugin in plugins:
    f = plugin.config_file()
    if f:
      _log.info('merging config of %s', plugin.id)
      # improved usage of method in context
      _c = _merge_config(_c, f, plugin.id)

  # override with more important settings
  global_ = os.path.abspath(os.environ.get('PHOVEA_CONFIG_PATH', 'config.json'))
  if os.path.exists(global_):
    with codecs.open(global_, 'r', 'utf-8') as fi:
      extend(_c, jsoncfg.loads(fi.read()))
  # merge changes done before the merge
  # check whether _pre_merge_changes is of NoneType
  if _pre_merge_changes is not None:
    extend(_c, _pre_merge_changes)
    _pre_merge_changes = None



###############################################################################
# Caleydo - Visualization for Molecular Biology - http://caleydo.org
# Copyright (c) The Caleydo Team. All rights reserved.
# Licensed under the new BSD license, available at http://caleydo.org/license
###############################################################################


from builtins import range
import json
from typing import Union


class JSONExtensibleEncoder(json.JSONEncoder):
  """
  json encoder with extension point extensions
  """

  def __init__(self, *args, **kwargs):
    super(JSONExtensibleEncoder, self).__init__(*args, **kwargs)

    from .plugin import list as list_plugins
    self.encoders = [p.load().factory() for p in list_plugins('json-encoder')]

  def default(self, o):
    for encoder in self.encoders:
      if o in encoder:
        return encoder(o, self)
    return super(JSONExtensibleEncoder, self).default(o)


def to_json(obj, *args, **kwargs):
  """
  convert the given object ot json using the extensible encoder
  :param obj:
  :param args:
  :param kwargs:
  :return:
  """
  if 'allow_nan' in kwargs:
    del kwargs['allow_nan']
  if 'indent' in kwargs:
    del kwargs['indent']
  kwargs['ensure_ascii'] = False

  # Pandas JSON module has been deprecated and removed. UJson cannot convert numpy arrays, so it cannot be used here. The JSON used here does not support the `double_precision` keyword.
  if isinstance(obj, float) or isinstance(obj, dict) or isinstance(obj, list):
    obj = _handle_nan_values(obj)
  return json.dumps(obj, cls=JSONExtensibleEncoder, *args, **kwargs)


def _handle_nan_values(obj_to_convert: Union[dict, list, float]) -> Union[dict, list, None]:
    """
    Convert any NaN values in the given object to None. Previously, Pandas was used to encode NaN to null. This feature has been deprecated and removed, therefore
    the standard JSON encoder is used which parses NaN instead of null. A custom JSON encoder does not work for converting these values to None because python's
    JSON encoder already knows how to serialize NaN values.
    :param obj_to_convert:
    :return dict, list or None:
    """
    import math
    converted_dict = {}
    converted_list = []
    # primitive value
    if (isinstance(obj_to_convert, float) and math.isnan(obj_to_convert)):
      return None
    # convert dictionaries
    if isinstance(obj_to_convert, dict):
      for k, v in obj_to_convert.items():
          # value is dictionary or list
          if isinstance(v, dict) or isinstance(v, list):
              converted_dict[k] = _handle_nan_values(v)
          else:
              # value is NaN
              if (isinstance(v, float) and math.isnan(v)):
                  converted_dict[k] = None
              else:
                  converted_dict[k] = v
      return converted_dict
    # convert lists
    elif isinstance(obj_to_convert, list):
      for elem in obj_to_convert:
        # list element is dictionary
        if isinstance(elem, dict):
          converted_list.append(_handle_nan_values(elem))
        # list element is NaN value
        elif (isinstance(elem, float) and math.isnan(elem)):
          converted_list.append(None)
        else:
          converted_list.append(elem)
      return converted_list


def jsonify(obj, *args, **kwargs):
  """
  similar to flask.jsonify but uses the extended json encoder and an arbitrary object
  :param obj:
  :param args:
  :param kwargs:
  :return:
  """
  from .ns import Response
  return Response(to_json(obj, *args, **kwargs), mimetype='application/json; charset=utf-8')


def glob_recursivly(path, match):
  import os
  import fnmatch

  for dirpath, dirnames, files in os.walk(path):
    if match is None:
      return None
    for f in fnmatch.filter(files, match):
      yield os.path.join(dirpath, f)


def fix_id(id):
  """
  fixes the id such that is it a resource identifier
  :param id:
  :return:
  """
  import re
  # convert strange characters to space
  r = re.sub(r"""[!#$%&'\(\)\*\+,\./:;<=>\?@\[\\\]\^`\{\|}~_]+""", ' ', id)
  # title case all words
  r = r.title()
  r = r[0].lower() + r[1:]
  # remove white spaces
  r = re.sub(r'\s+', '', r, flags=re.UNICODE)
  return r


def random_id(length):
  import string
  import random
  s = string.ascii_lowercase + string.digits
  id = ''
  for i in range(0, length):
    id += random.choice(s)
  return id



###############################################################################
# Caleydo - Visualization for Molecular Biology - http://caleydo.org
# Copyright (c) The Caleydo Team. All rights reserved.
# Licensed under the new BSD license, available at http://caleydo.org/license
###############################################################################


def parse_json(args, files):
  if 'desc' in args:
    import json
    args = json.loads(args['desc'])
  if 'nodes' not in args:
    args['nodes'] = []
  if 'edges' not in args:
    args['edges'] = []
  return args
